//thread pool 2311, see also 3vr.c

#include <cassert>
#include <iostream>
#include <vector>
#include <queue>

#include <algorithm>
#include <thread> //mutex function
#include <functional> //vs22
#include <mutex>
#include <condition_variable>
#include <chrono>
#include <future>

using std::vector;
using std::deque; 
using std::queue;

using std::cout;
using std::endl;
using std::thread;
using std::future;
using std::function;

using std::mutex;
using std::unique_lock;
using std::lock_guard;
using std::shared_ptr;
using std::condition_variable;
using std::invoke_result;
using std::packaged_task;
using std::bind;

#define WXC_PARALLEL //two threads make it faster
//#define WXC_NO_PTR_TASK //not good, messy
class CThreadPool
{
private:
	queue<function<void()>> mQueue; //binding is like to normalize the functions to standard form to be in q.
	mutex mMutex;
	condition_variable mCv;
	vector<thread> maThreads;
	bool mbShouldStop;
	void funcThread(int iThreadTag);
public:
	template<class F, class... Args> using RETURN_TYPE = typename invoke_result<F, Args...>::type;
	explicit CThreadPool(size_t iPoolSize);
	~CThreadPool();
	template<class F, class... Args> future< RETURN_TYPE<F, Args...> > Push(F&& f, Args&&... args);
	// should add some = delete; for all && and const &, ctor and assign
};
CThreadPool::CThreadPool(size_t iPoolSize)
{
	mbShouldStop = false;
	for (auto i = 0; i < iPoolSize; i++)
		maThreads.emplace_back(&CThreadPool::funcThread, this, i);
}
CThreadPool::~CThreadPool()
{
	{
		lock_guard<mutex> lock(mMutex);
		mbShouldStop = true;
	}
	mCv.notify_all();
	for (auto& thr : maThreads)
	{
		if(thr.joinable())
		thr.join();
	}
}
template<class F, class ...Args>
future<typename CThreadPool::RETURN_TYPE<F, Args...>> CThreadPool::Push(F&& f, Args && ...args)  //was inline 
{
	using return_type = RETURN_TYPE<F, Args...>;
#ifdef WXC_NO_PTR_TASK
	packaged_task<return_type()> *pTask = new
		packaged_task<return_type()>(bind(forward<F>(f), forward<Args>(args)...));
	future<return_type> fu = pTask->get_future();
	{
		lock_guard<mutex> lock(mMutex);
		mQueue.push([pTask]() { packaged_task<return_type()> task(move(*pTask));  task(); }); //task ctor called when mQueue out of scope
	}
#else
	shared_ptr< packaged_task<return_type()> > pTask =
		make_shared< packaged_task<return_type()> >(bind(forward<F>(f), forward<Args>(args)...));
	future<return_type> fu = pTask->get_future();
	{
		lock_guard<mutex> lock(mMutex);
		mQueue.emplace([pTask]() { (*pTask)(); });
	}
#endif
	mCv.notify_one();
	return fu;
}
void CThreadPool::funcThread(int iThreadTag)
{
	cout << "funcThread entered " << iThreadTag << endl;
	while (true)
	{
		function<void()> task;
		{
			unique_lock<mutex> lock(mMutex);
			while(!mbShouldStop && mQueue.empty())
				mCv.wait(lock, [this]() {
					return mbShouldStop || !mQueue.empty();
					});
			if (mbShouldStop && mQueue.empty()) {
				break;
			}
			if (mQueue.empty()) {
				continue;
			}
		#ifdef WXC_NO_PTR_TASK
			task = mQueue.front();
		#else
			task = mQueue.front();
		#endif
			mQueue.pop();
		}
		task();
	}
	cout << "funcThread exited " << iThreadTag << endl;
}
int funcTestThreadPool()
{
	cout << "funcTestThreadPool sleep 2 sec" << endl;
	sleep_for(seconds(2));
	cout << "funcTestThreadPool sleep done" << endl;
	return 3;
}
struct functorTestThreadPool
{
	int operator()(char)
	{
		cout << "functorTestThreadPool sleep 2 sec" << endl;
		sleep_for(seconds(2));
		cout << "functorTestThreadPool sleep done" << endl;
		return 2;
	}
};
void TestThreadPool()
{
#ifdef WXC_PARALLEL //8 is maxx on rioLe
	size_t iPoolSize = 2;
#else
	size_t iPoolSize = 1;
#endif
	CThreadPool myPool{ iPoolSize };
	functorTestThreadPool myFunctor;
	time_point<system_clock> now = system_clock::now();
	steady_clock::time_point begin = steady_clock::now();
	future<int> fuFunctor = myPool.Push(myFunctor, 'b');
	future<int> fuFunction = myPool.Push(funcTestThreadPool);
	fuFunctor.wait();
	int r = fuFunctor.get();
	assert(r == 2);
	cout << "TestThreadPool funcTestThreadPool will fuFunction.get" << endl;
	assert(fuFunction.get() == 3);
	cout << "TestThreadPool funcTestThreadPool done" << endl;
	steady_clock::time_point end = steady_clock::now();
	size_t d = duration_cast<seconds> (end - begin).count();
	cout << "Time  taken = " << d << " seconds" << " with pool size of " << iPoolSize << endl;
	if (iPoolSize >= 2)
		assert(d <= 2);
	if (iPoolSize == 1)
		assert(d >= 3);
	myPool.Push([]() {
		assert(funcTestThreadPool() == 3);
		});
}

//2402 parallel_for_eachUtil see 3vr.c
//#define WXC_THREAD //thread, otherwise asyn
//#define WXC_LAMBDA //task, otherwise regular function
#ifdef WXC_LAMBDA
#else
template<class Iterator, class Func> void parallel_for_eachFunc(Func func, Iterator begin, int iItemNum, int i)
{
	Iterator end = begin;
	std::advance(end, iItemNum);
	for_each(begin, end, func); //+ only for random access
	std::osyncstream(cout) << "parallel_for_eachFunc: " << "tag=" << i << " ID " << get_id() << " exiting" << endl;
}
#endif
template<class Iterator, class Func>
void parallel_for_eachUtil(Iterator Begin, Iterator End, Func func)
{
	int iThreadNum = 2;
	int iTotal = static_cast<int>(distance(Begin, End));
	int iItemPerTread = (iTotal + 1) / iThreadNum;
#ifdef WXC_THREAD
	vector<thread> aThreads; //just to prevent the threads dtor called
#else
	vector<future<void>> aFuture(iThreadNum);
#endif
	Iterator begin = Begin;
	for (int i = 0; i < iThreadNum; i++)
	{
		int iItemNum = (iItemPerTread * (i + 1) > iTotal) ? iTotal - iItemPerTread * i : iItemPerTread;
#ifdef WXC_LAMBDA
		packaged_task<void()> task([begin, func, iItemNum, i]() //cannot pass span which is good for vector not list
			{
				Iterator end = begin;
				std::advance(end, iItemNum);
				for_each(begin, end, func); //+ only for random access
				cout << "Thread: " << "tag=" << i << " ID " << get_id() << " exiting" << endl;
			}
		);
#endif
#ifdef WXC_THREAD
#ifdef WXC_LAMBDA
		aThreads.emplace_back(move(task));
#else
		aThreads.emplace_back(parallel_for_eachFunc<Iterator, Func>, func, begin, iItemNum, i);
#endif
#else
#ifdef WXC_LAMBDA
		aFuture[i] = async(move(task));
#else
		aFuture[i] = async(parallel_for_eachFunc<Iterator, Func>, func, begin, iItemNum, i);
#endif
#endif
		std::advance(begin, iItemNum);
	}
	for (int i = 0; i < iThreadNum; i++)
	{
#ifdef WXC_THREAD
		aThreads[i].join(); //still need join even after wait()?
#else
		aFuture[i].wait();
#endif

	}
	return;
}
void parallel_for_eachHelperDouble(int& input)
{
	input *= 2;
}
void parallel_for_eachTest()
{
	struct ST_DATA {
		list<int> maInput; //vector tested
		int miLastElement;
		function<void(int)> mFunc;
	};
	vector < ST_DATA > aTestData =
	{
		{
			{ 1,2,3, },
			6,
		},
		{
			{ 1,2,3,4 },
			8,
		},
	};
	size_t iSize = aTestData.size();
	cout << "parallel_for_eachTest" << endl;
	int iStart = 0;
	for (size_t i = iStart; i < iSize; i++)
	{
		cout << "started i = " << i << endl;
		parallel_for_eachUtil(aTestData[i].maInput.begin(), aTestData[i].maInput.end(), parallel_for_eachHelperDouble);
		decltype(aTestData[i].maInput)::iterator it = std::prev(aTestData[i].maInput.end());
		assert(*it == aTestData[i].miLastElement);
	}
	cout << " success" << endl;
}